#!/usr/bin/python2.7
###############################################################################
#@Copyright     Copyright (c) Imagination Technologies Ltd. All Rights Reserved
#@License       Strictly Confidential.
###############################################################################

import sys
import re
import argparse
import os
import atexit

import pdb
# Run in debugger> python -m pdb myscript.py, or add pdb.set_trace() as a breakpoint

reReplacedPMR = re.compile('^PMR(\d*)_X_(.*)$')
rePagePMR = re.compile('^PMR(\d*)_(\d*)_(.*)$')

def parse_arguments():
	parser = argparse.ArgumentParser(description="Given a PDump script file with -i, will produce a \
	PDump script where the page-wise MALLOCs of one PMR will be merged as much as possible into fewer commands. \
	Will rename the input file from \"input2.txt\" to \"input2.txt.old\" \
	and create an output with the name (path) given by the -o option or \"out2.txt\" as default. \
	 Example: \"python_optimise.py -i ../a/folder/out2.txt\" \
	 will move the script to ../a/folder/out2.txt.old and add a new file ../a/folder/out2.txt that has been processed.")

	# Mandatory
	parser.add_argument('-i',
					dest = "in_file",
					action = "store",
					help = "The PDump script input file to optimise.",
					metavar = "<Input File Path>",
					required = False,
					default = "out2.txt")

	parser.add_argument('-o',
					dest = "out_file",
					action = "store",
					help = "File name of the optimised script output. If only the file name is given the path from the input file will be used",
					metavar = "<Output File Path>",
					required = False,
					default = "not_given")

	parser.add_argument('-d',
					dest = "delete_in_file",
					action = "store_true",
					help = "Deletes the input file after processing",
					required = False,
					default = False)

	args = parser.parse_args()

	return args


def load_file(filename):
	f = open(filename, "r")
	return f.read().split("\n")

def save_file(filename, lines):
	f = open(filename, "w")
	for i in range(len(lines)):
		f.write(lines[i].strip() + "\n")
	print 'Saved ' + str(len(lines)) + ' lines'

def deconstruct_pmr_name(s):
	ents = s.split(':')

	if len(ents) > 1:
		memsp = ents[1]
	else:
		memsp = None

	if len(ents) > 2:
		pmrstr = ents[2]
	else:
		pmrstr = None

	if len(ents) > 3:
		offset = ents[3]
	else:
		offset = None

	if offset:
		offset = int(offset, 16)

	if (not pmrstr) or (pmrstr.find('PMR') != 0):
		return None

	# m = re.search('^PMR(\d*)_X_(.*)$', pmrstr);
	m = reReplacedPMR.search(pmrstr)

	if m:
		return {
			'memsp' : memsp,
			'serial' : m.group(1),
			'pgnum' : -1,
			'name' : m.group(2),
			'offset' : offset
		}

	# m = re.search('^PMR(\d*)_(\d*)_(.*)$', pmrstr)
	m = rePagePMR.search(pmrstr)

	# we MUST have a match here. if not then the PMR name is not in an
	# expected format!
	assert m

	return {
		'memsp' : memsp,
		'serial' : m.group(1),
		'pgnum' : int(m.group(2)),
		'name' : m.group(3),
		'offset' : offset
	}

# reduce the list to targets only
def filter_pmr_map(pmrmap):
	out = {}

	for key in pmrmap:
		pmr = pmrmap[key]

		if pmr['contiguous'] is True:
			out[key] = pmrmap[key]

	return out

def build_pmr_map(inlines):
	pmrs = {}
	linenum = 0

	for i in range(len(inlines)):
		line = inlines[i]
		linenum += 1
		ents = line.split(' ')

		if (ents[0] == 'MALLOC') or \
			(ents[0] == 'CALLOC'):
				pmrinfo = deconstruct_pmr_name(ents[1])

				#if linenum == 80:
				#	linenum = 80

				if pmrinfo:
					if not pmrinfo['serial'] in pmrs:
						pmrs[pmrinfo['serial']] = {}
						pmrs[pmrinfo['serial']]['pglist'] = [];
						# assume contiguous until we find otherwise
						pmrs[pmrinfo['serial']]['contiguous'] = True
						pmrs[pmrinfo['serial']]['pgsize'] = int(ents[2],0)

					if pmrinfo['pgnum'] != len(pmrs[pmrinfo['serial']]['pglist']):
						pmrs[pmrinfo['serial']]['contiguous'] = False;

					pmrs[pmrinfo['serial']]['pglist'].append(pmrinfo['pgnum'])

	return pmrs

def generate_pmr_name(pmrinfo):
	# Name length changes in per-page PDumps due to length restriction and growing serial number digits
	return 'PMR' + pmrinfo['serial'] + '_X_' + pmrinfo['name'][:45]

def int_to_hex(i):
	return '0x' + hex(i)[2:].upper()

def pdump_rewrite_malloc_calloc_free(pmrs, inlines):
	outlines = []
	linenum = 0

	for i in range(len(inlines)):
		line = inlines[i]
		linenum += 1

		ents = line.split(' ')
		copyline = True

		if ents[0] in ['MALLOC', 'CALLOC', 'FREE']:
			pmrinfo = deconstruct_pmr_name(ents[1])

			#if linenum == 80:
			#	linenum = 80

			if pmrinfo:
				if pmrinfo['serial'] in pmrs:
					pmr = pmrs[pmrinfo['serial']]
					# if this is the first malloc/calloc/free for a split/sparse allocation then output a single
					# malloc/calloc/free to cover the whole range
					if pmrinfo['pgnum'] == 0:
						s = None
						total_size = pmr['pgsize'] * len(pmr['pglist'])
						newname = generate_pmr_name(pmrinfo)

						if ents[0] == 'FREE':
							s = 'FREE :' + pmrinfo['memsp'] + ':' + newname
						elif ents[0] == 'MALLOC':
							s = 'MALLOC :' + pmrinfo['memsp'] + ':' + newname + ' ' + int_to_hex(total_size) + ' ' + ents[3]
						elif ents[0] == 'CALLOC':
							s = 'CALLOC :' + pmrinfo['memsp'] + ':' + newname + ' ' + int_to_hex(total_size) + ' ' + ents[3] + ' ' + ents[4]

						outlines.append(s)
					copyline = False
		if copyline:
			outlines.append(line)

	return outlines

def generate_ldb_map(inlines):
	ldbs = []
	state = {
		'in_ldb_group' : False,
		'startline' : 0,
		'pmrinfos' : [],
		'commands' : [],
	}

	for i in range(len(inlines) + 1):
		line = '   '

		# we fake an additional line in order to get an extra/final loop iteration
		# the extra "line" is "    "
		if i < len(inlines):
			line = inlines[i]

		ents = line.split(' ')

		is_ldb = ents[0] == 'LDB'
		is_cont = False
		pmrinfo = None
		cmd = None

		if is_ldb:
			pmrinfo = deconstruct_pmr_name(ents[1])
			if state['in_ldb_group']:
				is_cont = pmrinfo['serial'] == state['pmrinfos'][0]['serial']

			cmd = {
				'size' : int(ents[2], 16),
				'fileoffset' : int(ents[3], 16),
				'filename' : ents[4]
			}

		# if we are in a group of LDB commands and this line is not a
		# continuation then push the info about the group of LDB commands
		if state['in_ldb_group'] and not is_cont:
			# only add this LDB group if there is more than one LDB in it
			if len(state['pmrinfos']) > 1:
				ldbs.append({
					'startline' : state['startline'],
					'pmrinfos' : state['pmrinfos'],
					'commands' : state['commands']
					})

			state['in_ldb_group'] = False

		# if this line is a LDB command references a PMR we are interested in...
		if is_ldb and pmrinfo:
			if not state['in_ldb_group']:
				# start a new LDB group
				state['startline'] = i + 1;
				state['pmrinfos'] = [ pmrinfo ];
				state['in_ldb_group'] = True;
				state['commands'] = [ cmd ];
			else:
				# append to the existing LDB group
				state['pmrinfos'].append(pmrinfo);
				state['commands'].append(cmd);

	return ldbs

def generate_sab_map(inlines):
	sabs = []
	state = {
		'in_sab_group' : False,
		'startline' : 0,
		'pmrinfos' : [],
		'commands' : [],
	}

	for i in range(len(inlines) + 1):
		line = '   '

		# we fake an additional line in order to get an extra/final loop iteration
		# the extra "line" is "    "
		if i < len(inlines):
			line = inlines[i]

		ents = line.split(' ')

		is_sab = ents[0] == 'SAB'
		is_cont = False
		pmrinfo = None
		cmd = None

		if is_sab:
			pmrinfo = deconstruct_pmr_name(ents[1])
			if state['in_sab_group']:
				is_cont = pmrinfo['serial'] == state['pmrinfos'][0]['serial']

			if len(ents) == 4:
				cmd = {
					'size' : 1,
					'fileoffset' : int(ents[2], 16),
					'filename' : ents[3]
				}
			elif len(ents) == 5:
				cmd = {
					'size' : int(ents[2], 16),
					'fileoffset' : int(ents[3], 16),
					'filename' : ents[4]
				}
			else:
				raise SyntaxError("generate_sab_map(): unknown command length" )

		# if we are in a group of LDB commands and this line is not a
		# continuation then push the info about the group of LDB commands
		if state['in_sab_group'] and not is_cont:
			# only add this LDB group if there is more than one LDB in it
			if len(state['pmrinfos']) > 1:
				sabs.append({
					'startline' : state['startline'],
					'pmrinfos' : state['pmrinfos'],
					'commands' : state['commands']
					})

			state['in_sab_group'] = False

		# if this line is a LDB command references a PMR we are interested in...
		if is_sab and pmrinfo:
			if not state['in_sab_group']:
				# start a new LDB group
				state['startline'] = i + 1;
				state['pmrinfos'] = [ pmrinfo ];
				state['in_sab_group'] = True;
				state['commands'] = [ cmd ];
			else:
				# append to the existing LDB group
				state['pmrinfos'].append(pmrinfo);
				state['commands'].append(cmd);

	return sabs

def filter_ldb_map(pmrs, ldbs):
	out = []

	for i in range(len(ldbs)):
		group = ldbs[i]
		grouplen = len(group['commands'])
		keep = True

		assert len(group['commands']) == len(group['pmrinfos'])

		if grouplen == 1:
			# only one LDB in this group, so skip it.
			continue

		pmr_serialno = group['pmrinfos'][0]['serial']

		# check whether this PMR has been merged
		if pmr_serialno not in pmrs:
			keep = False

		if not keep:
			continue

		page_size = pmrs[pmr_serialno]['pgsize']

		for j in range(grouplen):
			# range starts with nonzero offset, so we cannot combine this.
			if group['pmrinfos'][j]['offset'] != 0:
				keep = False
				break

		if not keep:
			continue

		# check the page numbers are sequential
		for j in range(grouplen - 1):
			if group['pmrinfos'][j + 1]['pgnum'] != (group['pmrinfos'][j]['pgnum'] + 1):
				keep = False
				break

		if not keep:
			continue

		firstcmd = group['commands'][0];
		lastcmd = group['commands'][grouplen - 1];

		# check the total range is the expected size, given the number of LDBs
		# and the assumed page size
		if ((lastcmd['fileoffset'] - firstcmd['fileoffset']) / (grouplen - 1)) != page_size:
			keep = False
			continue

		if keep:
			out.append(group)

	return out

def filter_sab_map(pmrs, sabs):
	out = []

	for i in range(len(sabs)):
		group = sabs[i]
		grouplen = len(group['commands'])
		keep = True

		assert len(group['commands']) == len(group['pmrinfos'])

		if grouplen == 1:
			# only one SAB in this group, so skip it.
			continue

		pmr_serialno = group['pmrinfos'][0]['serial']

		# check whether this PMR has been merged
		if pmr_serialno not in pmrs:
			keep = False

		if not keep:
			continue

		page_size = pmrs[pmr_serialno]['pgsize']

		for j in range(grouplen):
			# range starts with nonzero offset, so we cannot combine this.
			if group['pmrinfos'][j]['offset'] != 0:
				keep = False
				break

		if not keep:
			continue

		# check the page numbers are sequential
		for j in range(grouplen - 1):
			if group['pmrinfos'][j + 1]['pgnum'] != (group['pmrinfos'][j]['pgnum'] + 1):
				keep = False
				break

		if not keep:
			continue

		firstcmd = group['commands'][0];
		lastcmd = group['commands'][grouplen - 1];

		# check the total range is the expected size, given the number of LDBs
		# and the assumed page size
		if ((lastcmd['fileoffset'] - firstcmd['fileoffset']) / (grouplen - 1)) != page_size:
			keep = False
			continue

		if keep:
			out.append(group)

	return out


def generate_ldb_command_for_group(pmrs, ldb):
	totbytes = 0

	pmr_serialno = ldb['pmrinfos'][0]['serial']
	page_size = pmrs[pmr_serialno]['pgsize']

	for i in range(len(ldb['commands'])):
		totbytes += ldb['commands'][i]['size']

	new_ldb_cmd = 'LDB :' + ldb['pmrinfos'][0]['memsp'] + ':' + generate_pmr_name(ldb['pmrinfos'][0]) + ':' + \
				int_to_hex(int(ldb['pmrinfos'][0]['pgnum'])*page_size) + \
				' '  + int_to_hex(totbytes) + ' ' + int_to_hex(ldb['commands'][0]['fileoffset']) + ' ' + \
				ldb['commands'][0]['filename']
	return new_ldb_cmd

def execute_ldb_map(pmrs, ldbs, inlines):
	outlines = []
	idx = 0

	for i in range(len(inlines)):
		line = inlines[i]
		linenumber = i + 1

		if idx < len(ldbs):
			curr = ldbs[idx]
			currlen = len(curr['commands'])

			if linenumber >= curr['startline']:
				if linenumber == curr['startline']:
					# emit the new LDB
					outlines.append(generate_ldb_command_for_group(pmrs, curr))
					continue
				elif linenumber < (curr['startline'] + currlen - 1):
					# just consume the original command
					continue
				elif linenumber == (curr['startline'] + currlen - 1):
					# this is the last original LDB command in the group,
					# so we now move on to the next LDB group
					idx += 1
					continue

		outlines.append(line)

	return outlines

def generate_sab_command_for_group(pmrs, sab):
	totbytes = 0

	pmr_serialno = sab['pmrinfos'][0]['serial']
	page_size = pmrs[pmr_serialno]['pgsize']

	for i in range(len(sab['commands'])):
		totbytes += sab['commands'][i]['size']

	new_sab_cmd = 'SAB :' + sab['pmrinfos'][0]['memsp'] + ':' + generate_pmr_name(sab['pmrinfos'][0]) + ':' + \
				int_to_hex(int(sab['pmrinfos'][0]['pgnum'])*page_size) + \
				' '  + int_to_hex(totbytes) + ' ' + int_to_hex(sab['commands'][0]['fileoffset']) + ' ' + \
				sab['commands'][0]['filename']

	return new_sab_cmd

def execute_sab_map(pmrs, sabs, inlines):
	outlines = []
	idx = 0

	for i in range(len(inlines)):
		line = inlines[i]
		linenumber = i + 1

		if idx < len(sabs):
			curr = sabs[idx]
			currlen = len(curr['commands'])

			if linenumber >= curr['startline']:
				if linenumber == curr['startline']:
					# emit the new LDB
					outlines.append(generate_sab_command_for_group(pmrs, curr))
					continue
				elif linenumber < (curr['startline'] + currlen - 1):
					# just consume the original command
					continue
				elif linenumber == (curr['startline'] + currlen - 1):
					# this is the last original LDB command in the group,
					# so we now move on to the next LDB group
					idx += 1
					continue

		outlines.append(line)

	return outlines

def rename_pmr_references(inlines, pmrs):
	outlines = []

	for i in range(len(inlines)):
		line = inlines[i]
		ents = line.split(' ');
		sout = ''

		if ents[0] == '--':
			sout = line
		else:
			for j in range(len(ents)):
				pmrinfo = deconstruct_pmr_name(ents[j])

				if pmrinfo and (pmrinfo['pgnum'] != -1) and pmrinfo['serial'] in pmrs:
					pmr = pmrs[pmrinfo['serial']]
					if pmr:

						sout += ':' + pmrinfo['memsp'] + ':'
						sout += generate_pmr_name(pmrinfo)

						offset = pmrinfo['pgnum'] * pmr['pgsize'] ;
						if pmrinfo['offset'] is not None:
							offset += pmrinfo['offset'];

						sout += ':' + int_to_hex(offset);
					else:
						sout += ents[j]
				else:
					sout += ents[j]

				# add a space between the parts of the command
				# but ensure we don't leave a trailing space
				if j != (len(ents) - 1):
					sout += ' '

		outlines.append(sout)

	return outlines

g_args = parse_arguments()


def main():
	optimisedToken = "-- OPTIMISED"
	loadfile = None
	infilename = g_args.in_file
	infilenameOld = infilename + ".old"
	i = 0
	while os.path.isfile(infilenameOld) is True:
		infilenameOld = infilename + ".old." + str(i)
		i += 1

	rename = None
	renamed = None

	# If the user did not supply an output file name the input file has to be
	# renamed and the output file name will be the previous input file name
	if g_args.out_file is "not_given":
		rename = True
		outfilename = infilename
	else:
		rename = False
		outfilename = g_args.out_file

	try:
		# Rename input file
		if rename == True:
			os.rename(infilename, infilenameOld)
			loadfile = infilenameOld
			renamed = True
		else:
			loadfile = infilename

		inlines = load_file(loadfile)

		# Check whether we already optimised this script and throw an error if yes.
		# The script will not work on an already optimised version
		if inlines[1] == optimisedToken:
			raise StandardError('Script already optimised, run on a clean version')
		else:
			inlines.insert(1, optimisedToken)

		#pdb.set_trace()

		print "Building PMR map...",
		pmrs = build_pmr_map(inlines)
		print len(pmrs)

		print "Filtering PMR map...",
		pmrs = filter_pmr_map(pmrs)
		print len(pmrs)

		onglines = inlines
		print "Number of input lines: ", len(onglines)

		print "Rewrite MALLOC/CALLOC/FREE commands for "+str(len(pmrs))+" PMRs...",
		onglines = pdump_rewrite_malloc_calloc_free(pmrs, onglines);
		print len(onglines)



		print "Generating LDB map...",
		ldbs = generate_ldb_map(onglines)
		print len(ldbs)

		print "Filtering LDB map...",
		ldbs = filter_ldb_map(pmrs, ldbs)
		print len(ldbs)

	 	print "Executing LDB map...",
	 	onglines = execute_ldb_map(pmrs, ldbs, onglines)
		print len(onglines)



	 	print "Generating SAB map...",
	 	sabs = generate_sab_map(onglines)
		print len(sabs)

	 	print "Filtering SAB map...",
		sabs = filter_sab_map(pmrs, sabs)
		print len(sabs)

	 	print "Executing SAB map...",
	 	onglines = execute_sab_map(pmrs, sabs, onglines)
		print len(onglines)

	 	print "Renaming all other PMR references...",
	 	onglines = rename_pmr_references(onglines, pmrs);
		print len(onglines)

		print "Saving output to " + outfilename
		save_file(outfilename, onglines)

	except:
		# If something went wrong, rename the input file to its original name
		if renamed is True:
			os.rename(infilenameOld, infilename)
		raise
	else:
		# Delete the input file if we were given the 'delete' argument
		if renamed is True:
			if g_args.delete_in_file is True:
				os.remove(infilenameOld)

main()
